# Simple support for organizing vcs branches as complex project
#
# Copyright (C) 2008, 2009 Alexander Belchenko <bialix AT ukr net>
# License: GNU GPL v.2

from bzrlib.lazy_import import lazy_import
lazy_import(globals(), """
import os
import shutil
import tempfile

from bzrlib import (
    branch,
    bzrdir,
    errors,
    osutils,
    trace,
    transport,
    urlutils,
    workingtree,
    )
from bzrlib.commands import run_bzr
from bzrlib.transport import local as _mod_local

from bzrlib.plugins import scmproj
from bzrlib.plugins.scmproj import (
    action,
    config,
    version_info,
    )
from bzrlib.plugins.scmproj.action import ActionList, url_for_display
from bzrlib.plugins.scmproj.vcs import BzrVcs, vcs_reg
""")


BASE       = '.scmproj'
CONFIG     = 'project.cfg'
ALTS_DIR   = 'alts'
ALTS_INDEX = 'index'
DEPOT      = 'box'

CONFIG_FN  = BASE + '/' + CONFIG    # candidate for removal

README_FILE = 'readme.txt'
README_TEXT = """\
This is a control directory for a project.

You need to have the bzr scmproj plugin installed
to be able work with the project itself.

See http://launchpad.net/bzr-scmproj
for more information about the scmproj plugin.
"""


class SilentStream(object):
    def write(self, s):
        pass
silent_file = SilentStream()


def is_url_local(url):
    return isinstance(transport.get_transport(url),
        _mod_local.LocalTransport)


def safe_path(name):
    """Encode name as safe cross-platform path for filesystem."""
    # this function implemented looking at hg store.py code
    # but we are using unicode instead of utf-8
    # so we can luxury to write it a bit shorter
    res = unicode(name).strip().translate(_SAFE_MAP)    # see end of the file
    parts = res.split('.', 1)
    prefix = parts[0]
    # special windows names (like aux or nul) should be mangled
    if prefix in _WINDOWS_RESERVED_FILENAMES:           # see end of the file
        encoded = prefix[:-1] + '~%02x' % ord(prefix[-1])
        parts[0] = encoded
        res = '.'.join(parts)
    if res[-1] == '.':
        res = res[:-1] + '~2e'     # 2e is hex code for dot
    return res


class NullConfig(object):
    """Special config to catch and report programming errors
    when we try to work with the project witout opened it first.
    It will raise BzrError on any access.
    """
    def __setattr__(self, k, v):
        raise errors.BzrError('Internal error: '
            'Project should be opened first')
    def __getattribute__(self, k):
        raise errors.BzrError('Internal error: '
            'Project should be opened first')


class CutLine(object):
    """Cut line helper.
    Print cut line to output stream starting from second call.
    """
    def __init__(self, stream, width=60):
        self.stream = stream
        self.line = '#%s\n' % ('-'*width)
        self.emit = False

    def __call__(self, force=False):
        if self.emit or force:
            self.stream.write(self.line)
        else:
            self.emit = True


class Project(object):

    def __new__(cls, url):
        if is_url_local(url):
           return super(Project, cls).__new__(ProjectLocal, url) 
        else:
           return super(Project, cls).__new__(ProjectRemote, url) 

    def __init__(self, url):
        """@param url: project root url."""
        self.branch = None
        self.cfg = NullConfig()
        self.settings = NullConfig()
        self.t = transport.get_transport(url)

    @staticmethod
    def _create_skeleton(path, branch_format=None):
        """Create .scmproj branch locally."""
        t = transport.get_transport(path)
        t.ensure_base()
        # create branch
        vo = BzrVcs()
        root = osutils.pathjoin(path, BASE)
        try:
            vo.create_branch(root, branch_format)
        except (errors.AlreadyBranchError,
                errors.BranchExistsWithoutWorkingTree):
            raise errors.BzrError(BASE+' directory already intialized.')
        # create mandatory files
        t = transport.get_transport(root)
        wt = workingtree.WorkingTree.open(root)
        t.put_bytes(CONFIG, config.Config().to_str())
        wt.add([CONFIG])
        t.mkdir(ALTS_DIR)
        wt.add([ALTS_DIR])
        index = ALTS_DIR + '/' + ALTS_INDEX
        t.put_bytes(index, '')
        wt.add([index])
        t.put_bytes(README_FILE, README_TEXT)
        wt.add([README_FILE])
        wt.commit('initialization')

    @staticmethod
    def initialize(url, branch_format=None):
        """Initialize new scmproj control dir."""
        t = transport.get_transport(url)
        if isinstance(t, _mod_local.LocalTransport):
            Project._create_skeleton(t.abspath('.'), branch_format)
        else:
            try:
                branch.Branch.open(url)
            except errors.NotBranchError:
                pass
            else:
                raise errors.BzrError('A branch already present at %s' % url)
            # create control dir in the temp dir and then push it to server
            temp_dir = tempfile.mkdtemp()
            Project._create_skeleton(temp_dir, branch_format)
            from_url = urlutils.join(urlutils.local_path_to_url(temp_dir), BASE)
            BzrVcs().push(url, from_url, use_existing_dir=True)
            try:
                shutil.rmtree(temp_dir)
            except (IOError, OSError), e:
                trace.mutter('Unable to delete temp directory %r: error: %s',
                    temp_dir, str(e))

    @staticmethod
    def _check_control_branch(branch):
        """The branch supplied should have project config committed.
        @raise: BzrError if this is plain branch.
        """
        branch.lock_read()
        try:
            rt = branch.repository.revision_tree(branch.last_revision())
            if rt.path2id('project.cfg') is None:
                raise errors.BzrError('Not scmproj control branch: %s' %
                    branch.base)
        finally:
            branch.unlock()

    def save_settings(self, alt=None, subset=None, workspace=None,
        override=False):
        """Save local settings values in setings config (branch.conf)."""
        if alt:
            if not self.settings.alt or override:
                self.settings.alt = alt
        if subset:
            if not self.settings.subset or override:
                self.settings.subset = subset
        if workspace:
            if not self.settings.workspace or override:
                self.settings.workspace = workspace

    def get_alt_name(self):
        """Return current alt name."""
        return self.settings.alt or self.cfg.get_alt_default_link()

    def get_subset_name(self):
        """Return current subset name."""
        return self.settings.subset or self.cfg.get_subset_default_link()

    def info(self, to_file=None):
        if to_file is None:
            import sys
            to_file = sys.stdout

        cfg = self.cfg
        settings = self.settings
        url = self.t.base

        to_file.write('Project: ')
        if cfg.name:
            to_file.write(cfg.name)
        else:
            to_file.write('unnamed')
        to_file.write('\n')
        to_file.write('URL: '+urlutils.unescape_for_display(url, 'utf-8')+'\n')
        if cfg.description:
            to_file.write('Description: '+cfg.description+'\n')
        if settings.project_url not in ('.', '', None):
            to_file.write('Parent URL: '+urlutils.unescape_for_display(
                settings.project_url, 'utf-8')+'\n')
        to_file.write('\n')
        if settings.alt:
            to_file.write('Alt: %s\n' % settings.alt)
        if settings.subset:
            to_file.write('Subset: %s\n' % settings.subset)

    def fetch(self, alt_name, to_file=None, component_list=None, all_branches=False):
        action_list = ActionList([action.FetchKBranchAction()])
        return self.run_actions(action_list, to_file, alt_name,
            component_list, all_branches)

    def fetch_from_source_branch(self, component, source_branch, to_file=None,
        alt_name=None):
        """Fetch one branch of one component from non-standard location."""
        action_list = ActionList([action.FetchFromBranchOnceAction(source_branch)])
        return self.run_actions(action_list, to_file, alt_name, [component])

    def run_cmds(self, command_list, to_file=None, ignore_errors=False,
        alt_name=None, components=None, cd_relpath=False, verbose=True,
        ):
        action_list = ActionList([action.BzrCommandsAction(command_list,
            ignore_errors, verbose)])
        return self.run_actions(action_list, to_file, alt_name,
            components, cd_relpath=cd_relpath, verbose=verbose)

    def run_actions(self, action_list, to_file=None, alt_name=None,
        component_list=None, all_branches=False, cd_relpath=False,
        verbose=True, workspace=None):
        """Generic iterator over component branches to execute some actions.

        @param  action_list:    list of Action instances to execute.
        @param  to_file:        stream to write messages.
        @param  component_list: optional list of components to iterate.
        @param  alt_name:       name of the alt to select component branches.
        @param  all_branches:   run actions for every branch of component;
                                this option override alt_name.
        @param  cd_relpath:     execute action in component working copy
                                (cd {RELPATH} before execution).
        @param  verbose:        show verbose info about actions

        @return: None if actions executed successfully for all components,
            otherwise return code will be used as return code for CLI.
        """
        if to_file is None:
            to_file = silent_file
        cutline = CutLine(to_file)
        if alt_name is None:
            alt_name = self.get_alt_name()
        alt_dict = self.cfg.get_alt(alt_name)
        if component_list is None:
            component_list = self.cfg.get_full_set()
        else:
            self.cfg.ensure_components(component_list)
        if workspace is None:
            workspace = self.settings.workspace or 'plain'
        if all_branches and workspace != 'boxed':
            raise errors.BzrError('Unable to run actions for all branches '
                'when workspace is not boxed.')

        project_root = urlutils.local_path_from_url(self.t.base)
        local_repos_base = osutils.pathjoin(project_root, BASE, DEPOT, 'repos')

        # local_paths dictionary
        # there should be paths in context of current component/kbranch
        local_paths = {
            'PROJECT_ROOT': project_root,
            'LOCAL_DEPOT': local_repos_base,
            }
        local_urls = {
            'PROJECT_ROOT': self.t.base,
            'LOCAL_DEPOT': urlutils.local_path_to_url(local_repos_base),
            }
        orig_cwd = os.getcwdu()
        # execute prolog
        for func in getattr(action_list, 'prolog', []):
            cutline()
            func()
        # let's go
        for kname in component_list:
            cutline()
            to_file.write('# Component: %s\n' % kname)
            ko = Component(self.cfg, kname)

            local_tree = osutils.pathjoin(project_root, ko.relpath)
            tree_url = urlutils.local_path_to_url(local_tree)
            local_paths['LOCAL_TREE'] = local_tree
            local_urls['LOCAL_TREE'] = tree_url
            if workspace != 'boxed':
                local_paths['LOCAL_BRANCH'] = local_tree
                local_urls['LOCAL_BRANCH'] = tree_url

            depot_repo = osutils.pathjoin(local_repos_base, safe_path(kname))
            repo_url = urlutils.local_path_to_url(depot_repo)
            local_paths['DEPOT_REPO'] = depot_repo
            local_urls['DEPOT_REPO'] = repo_url
            if workspace == 'boxed':
                local_paths['LOCAL_REPO'] = depot_repo
                local_urls['LOCAL_REPO'] = repo_url

            if all_branches:
                branches = ko.branches
            else:
                bname = alt_dict[kname]
                if bname is None:       # this component is missed in the alt
                    continue
                branches = [bname]
            if cd_relpath and ko.relpath != '.':
                # XXX this command may raise OSError if ko.relpath does not exist yet
                #     we should catch this error and report it as BzrError
                # TODO: show the relpath in the actions
                if verbose:
                    to_file.write("#  Entering directory `%s'\n" % ko.relpath)
                os.chdir(local_tree)
            try:
                for bname in branches:
                    kb = ko.get_branch(bname)
                    depot_branch = osutils.pathjoin(depot_repo, safe_path(kb.name))
                    branch_url = urlutils.local_path_to_url(depot_branch)
                    local_paths['DEPOT_BRANCH'] = depot_branch
                    local_urls['DEPOT_BRANCH'] = branch_url
                    if workspace == 'boxed':
                        local_paths['LOCAL_BRANCH'] = depot_branch
                        local_urls['LOCAL_BRANCH'] = branch_url
                    # we ready to run actions for this branch
                    for action in action_list:
                        ret_code = action(to_file, ko, kb, local_paths, local_urls)
                        if ret_code is not None:
                            return ret_code
            finally:
                if cd_relpath:
                    os.chdir(orig_cwd)
        # execute epilog
        for func in getattr(action_list, 'epilog', []):
            cutline()
            func()
        cutline()
        to_file.write('# Done.\n')

    def create_working_copy(self, to_file=None, alt_name=None,
        component_list=None, fetch=False, update=False, workspace=None,
        components_only=False):
        """Create/update boxed workspace."""
        action_list = ActionList()
        if fetch:
            action_list.append(action.FetchKBranchAction())
        action_list.append(action.CreateWorkingCopyAction(update))
        return self.run_actions(action_list, to_file, alt_name,
            component_list, workspace=workspace)

    def update_workspace(self, to_file=None, alt_name=None,
        component_list=None, workspace=None, update=True,
        components_only=False):
        """Update project workspace: create and update component trees."""
        if workspace == 'boxed':
            return self.create_working_copy(to_file, alt_name,
                component_list, fetch=True, update=update,
                workspace=workspace)
        # plain workspace
        action_list = ActionList([
            action.UpdateWorkingTreeAction(create=True),
            ])
        return self.run_actions(action_list, to_file, alt_name,
            component_list, workspace=workspace)

    def update_control_branch(self, to_file=None):
        """Special pre-action to update control branch with config.
        Should be executed before any other project actions
        because it may re-read config.
        """
        if to_file is None:
            to_file = silent_file
        from_url = self.branch.get_parent()
        if from_url:
            to_file.write('# Updating control branch from %s...\n' % (
                action.url_for_display(from_url)))
            vo = BzrVcs()
            vo.pull(self.branch.base, from_url)
            # project config could be changed by this action, need to re-read it
            self.open()
        else:
            to_file.write('# Skip control branch: no parent location known.\n')
        CutLine(to_file)(force=True)

    def commit(self, msgfile, to_file=None, component_list=None,
        components_only=False):
        """Commit the project: changes in components and control branch."""
        action_list = ActionList([
            action.CommitAction(osutils.abspath(msgfile)),
            ])
        if not components_only:
            action_list.epilog = [
                self._get_commit_control_branch_action(msgfile, to_file)]
        return self.run_actions(action_list, to_file,
            component_list=component_list)

    def _get_commit_control_branch_action(self, msgfile, to_file):
        def commit_control_branch():
            vo = BzrVcs()
            path = urlutils.local_path_from_url(self.branch.base)
            if vo.is_tree_changed(path):
                vo.commit_tree(path, msgfile)
        return commit_control_branch

    def get_component_list(self, subset=None, full_set=None,
        includes=None, excludes=None):
        """Return component list based on subset.
        If neither subset nor full_set specified then components of current
        project subset is returned.
        """
        if full_set:
            subset = 'FULL'
        elif subset is None:
            subset = self.get_subset_name()
        component_list = self.cfg.get_subset(subset)
        if includes or excludes:
            r = set(component_list)
            all = set(self.cfg.components)
            if includes:
                for i in includes:
                    if i in all:
                        r.add(i)
                    else:
                        raise errors.BzrError('Unknown component: '+i)
            if excludes:
                for e in excludes:
                    if e in all:
                        if e in r:
                            r.remove(e)
                    else:
                        raise errors.BzrError('Unknown component: '+e)
            component_list = self.cfg.sort_components(r)
        return component_list

    def cli_helper(self, alt=None, subset=None, full_set=None,
        includes=None, excludes=None, workspace=None):
        """@return: (alt, component_list, workspace)."""
        if alt is None:
            alt = self.get_alt_name()
        component_list = self.get_component_list(subset, full_set,
            includes, excludes)
        if workspace is None:
            workspace = self.settings.workspace or 'plain'
        return (alt, component_list, workspace)

    def publish(self, to_file=None, alt_name=None,
        component_list=None, all_branches=False, workspace=None,
        components_only=False):
        """Push changes from local component branches to their origins."""
        action_list = ActionList([action.PublishKBranch()])
        if not components_only:
            action_list.epilog = [
                self._get_publish_control_branch_action(to_file)]
        return self.run_actions(action_list, to_file, alt_name, component_list,
            all_branches, workspace=workspace)

    def _get_publish_control_branch_action(self, to_file=None):
        def publish_control_branch():
            to_url = self.branch.get_push_location()
            if to_url:
                to_file.write('# Push control branch to %s...\n' % (
                    action.url_for_display(to_url)))
                vo = BzrVcs()
                vo.push(to_url, self.branch.base)
            else:
                to_file.write('# Skip control branch: no push location known.\n')
        return publish_control_branch

    @staticmethod
    def _clone_to_local_url(br_from, to_url, accelerator_tree=None):
        """Internal machinery of clone (local,remote) -> local.
        @param  br_from:    source branch instance
        @param  to_url:     destination local URL
        @param  accelerator_tree:   can be used for local->local clone.
        """
        br_from.lock_read()
        try:
            Project._check_control_branch(br_from)
            # create base dir if needed
            transport.get_transport(to_url).ensure_base()
            to_ctrl_url = urlutils.join(to_url, BASE)
            try:
                transport.get_transport(to_ctrl_url).mkdir('.')
            except errors.FileExists:
                raise errors.BzrError('Target directory "%s" already'
                    ' exists.' % url_for_display(to_ctrl_url))
            br_from.bzrdir.sprout(to_ctrl_url,
                force_new_repo=True, # forces it to create a working tree,
                # I don't think we want our control branch in a repo.
                accelerator_tree=accelerator_tree, source_branch=br_from)
        finally:
            br_from.unlock()


class ProjectLocal(Project):
    """Local copy of the Project."""

    def clone(self, to_url):
        """clone local->local."""
        from_url = urlutils.join(self.t.base, BASE)
        accelerator_tree, br_from = bzrdir.BzrDir.open_tree_or_branch(from_url)
        Project._clone_to_local_url(br_from, to_url, accelerator_tree)

    def open(self):
        """Open project for business.

        @return: self.
        """
        self.branch = branch.Branch.open(self.t.abspath(BASE))
        # read and parse project.cfg
        self.cfg = config.Config(self.t.get_bytes(CONFIG_FN))
        self.settings = config.LocalSettings(self.branch)
        # check for the presense of subprojects
        if self.cfg.subprojects:
            trace.warning("The scmproj plugin version %s "
                "(from %s) does not support subprojects yet, "
                "but your config file has declared subprojects.",
                '.'.join([str(i) for i in scmproj.version_info]),
                os.path.abspath(os.path.dirname(__file__)))
        return self


class ProjectRemote(Project):
    """Project as the branch on the remote server."""

    def clone(self, to_url):
        """clone remote->local."""
        from_url = self.t.base
        br_from = branch.Branch.open(from_url)
        Project._clone_to_local_url(br_from, to_url)

    def open(self):
        raise errors.BzrError("Remote project cannot be opened.")


class Component(object):
    """Project component."""

    def __init__(self, config_object, name):
        """Create component object based on data from project.cfg.
        @param  config_object:  Config (project.cfg wrapper) instance
        @param  name:   name of the component
        """
        self.name = name
        self._cfg = config_object
        self._data = config_object.get_component_section(name)
        self.description = self._data.get('DESCRIPTION')
        self.vcs = self._data.get('VCS', config_object.get_default_vcs())
        self.format = (self._data.get('FORMAT')
            or config_object.get_default_format())
        self.branches = self._cfg.get_component_branches(self.name)

    def _get_relpath(self):
        return self._cfg.get_component_relpath(self.name)

    relpath = property(_get_relpath)

    def get_branch(self, branch_name='DEFAULT', alt_name='DEFAULT'):
        """@return: KBranch object"""
        if branch_name == 'DEFAULT':
            branch_name = self.branches[0]
        if branch_name in self.branches:
            sb = self._cfg.get_branch_section(self.name, branch_name)
            if sb is None:
                # simple branch
                return KBranch(self._cfg, self.name, branch_name,
                    self.vcs, format=self.format)
            else:
                # explicitly defined branch
                vcs = sb.get('VCS', self.vcs)
                format = sb.get('FORMAT', self.format)
                revision = sb.get('REVISION')
                extra = sb
                return KBranch(self._cfg, self.name, branch_name,
                    vcs,revision=revision,
                    format=format, extra=extra)
        raise KeyError


class KBranch(object):
    """Component branch."""

    def __init__(self, cfg, component_name, name,
            vcs, revision=None, format=None, extra=None):
        self._cfg = cfg
        self.component_name = component_name
        self.name = name
        self.vcs = vcs
        self.format = format
        self.revision = revision
        self.extra = extra or {}

    def _get_url(self):
        return self._cfg.get_component_branch_url(self.component_name,
            self.name)

    url = property(_get_url)


class Snapshot(object):
    """Project snapshot."""

    def __init__(self, project):
        if project.formats and project.formats.get('snapshot'):
            self.project = project
            self.vcs = project.formats['snapshot']
            self.url = urlutils.join(project.t.base, SNAPSHOT_DIR)
            self._path = None
        else:
            raise errors.BzrError('project without snapshot')

    def _get_path(self):
        """Return local path to snapshot branch/tree."""
        if self._path is None:
            self._path = urlutils.local_path_from_url(self.url)
        return self._path

    path = property(_get_path)

    def collect_revids(self):
        act = action._CollectRevidsAction()
        ret = self.project.run_actions([act])
        if ret not in (None, 0):
            raise errors.BzrError('Collect revids failed with return code %r' % ret)
        return act.components, act.subprojects

    def update_snapshot(self, cfg=True):
        """Update snapshot data: index card and project.cfg
        @param  cfg:    if True then update project.cfg
        """
        if cfg:
            t = self.project.t
            s = t.get_bytes(CONFIG_FN)
            dest = urlutils.join(SNAPSHOT_DIR, CONFIG_BASE)
            t.put_bytes(dest, s)
            s_vo = vcs_reg.get(self.vcs)()
            s_vo.add(self.path, CONFIG_BASE)
        self.update_index_card()

    def update_index_card(self):
        components, subprojects = self.collect_revids()
        ic = self.get_index_card()
        if ic is None:
            ic = config.IndexCard()
        ic.alt = self.project.get_alt_name()
        ic.subset = self.project.get_subset_name()
        # TODO we should re-sort all components/subprojects
        #      in alphabetical order, not just append new ones
        for k in sorted(components.keys()):
            ic.set_component(k, components[k])
        for s in sorted(subprojects.keys()):
            ic.set_subproject(s, subprojects[s])
        self.set_index_card(ic)

    def get_index_card(self):
        t = self.project.t
        card_url = urlutils.join(SNAPSHOT_DIR, self.project.get_alt_name())
        if t.has(card_url):
            return config.IndexCard(t.get_bytes(card_url))
        return None

    def set_index_card(self, card):
        t = self.project.t
        card_url = urlutils.join(SNAPSHOT_DIR, self.project.get_alt_name())
        t.put_bytes(card_url, card.to_str())
        s_vo = vcs_reg.get(self.vcs)()
        s_vo.add(self.path, self.project.get_alt_name())

    def commit_snapshot(self, msgfile):
        s_vo = vcs_reg.get(self.vcs)()
        if s_vo.is_tree_changed(self.url):
            return s_vo.commit_tree(self.path, msgfile, quiet=True)


#-----------------------------------------------------------------------------
# uninteresting internal stuff

_WINDOWS_RESERVED_FILENAMES = set('''con prn aux nul
    com1 com2 com3 com4 com5 com6 com7 com8 com9
    lpt1 lpt2 lpt3 lpt4 lpt5 lpt6 lpt7 lpt8 lpt9'''.split())

_SAFE_MAP = {}

def _create_safe_map():
    global _SAFE_MAP
    _SAFE_MAP.update(dict((i, u'~%02x' % i) for i in (
        ord(s) for s in ' \\:*?"<>|/~')))                       # reserved characters
    _SAFE_MAP.update(dict((i, u'_'+s.lower()) for i, s in
        ((i, chr(i)) for i in range(ord('A'), ord('Z')+1))))    # lower the caps lock
    _SAFE_MAP.update({ord('_'): u'__'})                         # special case of caps lock

_create_safe_map()
