# Simple support for organizing bzr branches as complex project
#
# Copyright (C) 2008, 2009 Alexander Belchenko <bialix AT ukr net>
# License: GNU GPL v.2

# TODO: split it to several files and put them into tests/ subdir.

from cStringIO import StringIO
import os

from bzrlib import (
    branch,
    commands,
    errors,
    trace,
    urlutils,
    )
from bzrlib.tests import TestCase, TestCaseWithTransport
from bzrlib.tests.test_sftp_transport import TestCaseWithSFTPServer
from bzrlib.transport import get_transport

from bzrlib.plugins.scmproj import (
    project,
    )
from bzrlib.plugins.scmproj.action import Action

from bzrlib.plugins.scmproj.project import (
    Component,
    Project,
    ProjectLocal,
    ProjectRemote,
    is_url_local,
    safe_path,
    )
from bzrlib.plugins.scmproj.config import (
    CONFIG_TMPL,
    Config,
    )
from bzrlib.plugins.scmproj.tests.whiteutils import reindentBlock

INSTALL_CMD = getattr(commands, "install_bzr_command_hooks", lambda: None)


class TestComponent(TestCase):

    def test_init(self):
        cfg = Config(reindentBlock("""\
            [component "foo"]
            RELPATH = .
            BRANCHES = trunk bar
            
                [[branch "bar"]]
                BRANCH_URL = http://another.net
                REVISION = xyz
            """))
        ko = Component(cfg, 'foo')
        self.assertEqual(None, ko.description)
        self.assertEqual('bzr', ko.vcs)
        self.assertEqual('.', ko.relpath)
        self.assertEqual(['trunk', 'bar'], ko.branches)
        # check optional values
        cfg = Config(reindentBlock("""\
            [component "foo"]
            DESCRIPTION = Mega component!
            RELPATH = .
            VCS = rcs
            BRANCHES = trunk bar
            
                [[branch "bar"]]
                BRANCH_URL = http://another.net
                REVISION = xyz
            """))
        ko = Component(cfg, 'foo')
        self.assertEqual('Mega component!', ko.description)
        self.assertEqual('rcs', ko.vcs)
        self.assertEqual('.', ko.relpath)
        self.assertEqual(['trunk', 'bar'], ko.branches)
        # mixed config
        cfg = Config(reindentBlock("""\
            [PROJECT]
            COMPONENTS = foo bar

            [COMPONENTS]
            RELPATH = {COMPONENT}
            BRANCH_URL = http://example.com/{COMPONENT}/trunk

            [component bar]
            RELPATH = .
            BRANCHES = dev
            """))
        ko = Component(cfg, 'foo')
        self.assertEqual(None, ko.description)
        self.assertEqual('bzr', ko.vcs)
        self.assertEqual('foo', ko.relpath)
        self.assertEqual(['trunk'], ko.branches)
        ko = Component(cfg, 'bar')
        self.assertEqual(None, ko.description)
        self.assertEqual('bzr', ko.vcs)
        self.assertEqual('.', ko.relpath)
        self.assertEqual(['dev'], ko.branches)

    def test_get_component_from_config(self):
        cfg = Config(reindentBlock("""\
            [component "foo"]
            RELPATH = .
            BRANCHES =
            """))
        self.assertEqual('.',
            Component(cfg, 'foo').relpath)

    def test_get_branch(self):
        cfg = Config(reindentBlock("""\
            [COMPONENTS]
            BRANCH_URL = http://example.com/{BRANCH}
            
            [component "foo"]
            RELPATH = .
            BRANCHES = trunk bar
            
                [[branch "bar"]]
                BRANCH_URL = http://another.net/spam
                REVISION = xyz
            """))
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('DEFAULT')
        self.assertEqual('trunk', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('trunk')
        self.assertEqual('trunk', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('bar')
        self.assertEqual('bar', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://another.net/spam', kb.url)
        self.assertEqual('xyz', kb.revision)
        ##
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('DEFAULT')
        self.assertEqual('bzr', kb.vcs)

        kb = ko.get_branch('bar')
        self.assertEqual('bzr', kb.vcs)

        cfg = Config(reindentBlock("""\
            [COMPONENTS]
            BRANCH_URL = http://example.com/{BRANCH}
            
            [component "foo"]
            DESCRIPTION = Mega component!
            RELPATH = .
            VCS = rcs
            BRANCHES = trunk bar baz
            
                [[branch "bar"]]
                BRANCH_URL = http://another.net/spam
                REVISION = xyz
            
                [[branch "baz"]]
                VCS = svn
                BRANCH_URL = https://svn.host.org/project
                REVISION = 1234
            """))
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('DEFAULT')
        self.assertEqual('rcs', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('bar')
        self.assertEqual('rcs', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://another.net/spam', kb.url)
        self.assertEqual('xyz', kb.revision)

        kb = ko.get_branch('baz')
        self.assertEqual('svn', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('https://svn.host.org/project', kb.url)
        self.assertEqual('1234', kb.revision)

        # optional format for branches
        cfg = Config(reindentBlock("""\
            [component "foo"]
            VCS = bzr
            FORMAT = pack-0.92
            RELPATH = .
            BRANCHES = trunk bar baz
            
                [[branch "bar"]]
                VCS = bzr
                FORMAT = rich-root-pack
                BRANCH_URL = http://another.net/spam
                REVISION = xyz
            
                [[branch "baz"]]
                BRANCH_URL = https://host.org/project
            """))
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('trunk')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual('pack-0.92', kb.format)

        kb = ko.get_branch('bar')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual('rich-root-pack', kb.format)

        kb = ko.get_branch('baz')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual('pack-0.92', kb.format)

        # standalone branch (not in shared repo) as '.'
        cfg = Config(reindentBlock("""\
            [COMPONENTS]
            BRANCH_URL = http://example.com/{COMPONENT}
            
            [component "foo"]
            RELPATH = .
            BRANCHES = .
            """))
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('DEFAULT')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/foo', kb.url)
        self.assertEqual(None, kb.revision)

        # component URL might be empty, but branch URL should not
        cfg = Config(reindentBlock("""\
            [component "foo"]
            RELPATH = .
            BRANCHES = spam eggs
            
                [[branch "eggs"]]
                REVISION = -1
            """))
        ko = Component(cfg, 'foo')

        self.assertRaises(errors.BzrError, ko.get_branch('spam')._get_url)
        self.assertRaises(errors.BzrError, ko.get_branch('eggs')._get_url)

        # mixed config
        cfg = Config(reindentBlock("""\
            [PROJECT]
            COMPONENTS = foo bar

            [COMPONENTS]
            RELPATH = {COMPONENT}
            BRANCH_URL = http://example.com/{COMPONENT}/trunk

            [component bar]
            RELPATH = .
            BRANCHES = dev
            """))
        ko = Component(cfg, 'foo')
        kb = ko.get_branch('DEFAULT')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/foo/trunk', kb.url)
        self.assertEqual(None, kb.revision)
        kb = ko.get_branch('trunk')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/foo/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        ko = Component(cfg, 'bar')
        kb = ko.get_branch('DEFAULT')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/bar/trunk', kb.url)
        self.assertEqual(None, kb.revision)
        kb = ko.get_branch('dev')
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/bar/trunk', kb.url)
        self.assertEqual(None, kb.revision)

    def test_branch_format_from_components_section(self):
        cfg = Config(reindentBlock("""\
            [COMPONENTS]
            BRANCH_URL = http://example.com/{BRANCH}
            FORMAT = rich-root-pack
            
            [component "foo"]
            RELPATH = .
            BRANCHES = trunk bar
            
                [[branch "bar"]]
                BRANCH_URL = http://another.net/spam
                REVISION = xyz
            """))
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('DEFAULT')
        self.assertEqual('trunk', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual('rich-root-pack', kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('trunk')
        self.assertEqual('trunk', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual('rich-root-pack', kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('bar')
        self.assertEqual('bar', kb.name)
        self.assertEqual('bzr', kb.vcs)
        self.assertEqual('rich-root-pack', kb.format)
        self.assertEqual('http://another.net/spam', kb.url)
        self.assertEqual('xyz', kb.revision)

    def test_branch_vcs_from_components_section(self):
        cfg = Config(reindentBlock("""\
            [COMPONENTS]
            BRANCH_URL = http://example.com/{BRANCH}
            VCS = lokum

            [component "foo"]
            RELPATH = .
            BRANCHES = trunk bar

                [[branch "bar"]]
                BRANCH_URL = http://another.net/spam
                REVISION = xyz
            """))
        ko = Component(cfg, 'foo')

        kb = ko.get_branch('DEFAULT')
        self.assertEqual('trunk', kb.name)
        self.assertEqual('lokum', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('trunk')
        self.assertEqual('trunk', kb.name)
        self.assertEqual('lokum', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://example.com/trunk', kb.url)
        self.assertEqual(None, kb.revision)

        kb = ko.get_branch('bar')
        self.assertEqual('bar', kb.name)
        self.assertEqual('lokum', kb.vcs)
        self.assertEqual(None, kb.format)
        self.assertEqual('http://another.net/spam', kb.url)
        self.assertEqual('xyz', kb.revision)

    def test_branch_url_for_component(self):
        # BRANCH_URL template could be defined in the component section
        cfg = Config(reindentBlock("""\
            [component "foo"]
            RELPATH = .
            BRANCHES = trunk dev release
            BRANCH_URL = http://example.com/{COMPONENT}/{BRANCH}

                [[branch "release"]]
                BRANCH_URL = http://example.com/releases/{COMPONENT}
            """))
        ko = Component(cfg, 'foo')
        self.assertEqual('http://example.com/foo/trunk',
            ko.get_branch('DEFAULT').url)
        self.assertEqual('http://example.com/foo/trunk',
            ko.get_branch('trunk').url)
        self.assertEqual('http://example.com/foo/dev',
            ko.get_branch('dev').url)
        self.assertEqual('http://example.com/releases/foo',
            ko.get_branch('release').url)

    def test_component_and_branch_vcs(self):
        cfg = Config(reindentBlock("""\
            [PROJECT]
            VCS = bzr

            [component "foo"]
            RELPATH = .
            BRANCHES = trunk
            BRANCH_URL = http://example.com/{COMPONENT}/{BRANCH}
            """))
        ko = Component(cfg, 'foo')
        self.assertEqual('bzr', ko.vcs)
        self.assertEqual('bzr', ko.get_branch('trunk').vcs)

    def test_branch_url_as_win32_path(self):
        cfg = Config(reindentBlock(
            r"""[PROJECT]
            VCS = bzr

            [component "foo"]
            RELPATH = .
            BRANCHES = trunk
            BRANCH_URL = C:\work\test\branch
            """))
        self.assertEqual("C:\\work\\test\\branch",
            Component(cfg, 'foo').get_branch('trunk').url)


class Test42(TestCase):

    def test_is_url_local(self):
        self.assertTrue(is_url_local('.'))
        self.assertTrue(is_url_local(urlutils.local_path_to_url('.')))
        self.assertFalse(is_url_local('http://example.com'))

    def test_project_constructor(self):
        self.assertTrue(isinstance(Project('.'), ProjectLocal))
        self.assertTrue(isinstance(Project('http://example.com'),
            ProjectRemote))

    def test_safe_path(self):
        self.assertEqual('foo', safe_path('foo'))
        self.assertEqual('_foo', safe_path('Foo'))
        self.assertEqual('__foo', safe_path('_foo'))
        self.assertEqual('foo~2fbar', safe_path('foo/bar'))
        self.assertEqual('foo~5cbar', safe_path('foo\\bar'))
        self.assertEqual('foo~7e', safe_path('foo~'))
        self.assertEqual('foo~2e', safe_path('foo.'))
        self.assertEqual('nu~6c.foo', safe_path('nul.foo'))
        self.assertEqual('com~31.foo', safe_path('com1.foo'))


class TestProject(TestCaseWithTransport):

    def setUp(self):
        TestCaseWithTransport.setUp(self)
        INSTALL_CMD()

    def test__create_skeleton(self):
        Project._create_skeleton('.')
        t = get_transport('.')
        self.assertIsDirectory('.scmproj', t)
        self.assertIsDirectory('.scmproj/.bzr', t)
        self.assertIsDirectory('.scmproj/alts', t)
        self.failUnlessExists([
            '.scmproj/project.cfg',
            '.scmproj/readme.txt',
            ])
        # check we have valid branch there
        br = branch.Branch.open('.scmproj')
        # check the revision tree has project config in it
        br.lock_read()
        try:
            rt = br.repository.revision_tree(br.last_revision())
            self.assertNotEqual(None, rt.path2id('project.cfg'))
        finally:
            br.unlock()

    def test__check_control_branch(self):
        wt = self.make_branch_and_tree('.')
        self.build_tree(['project.cfg'])
        wt.add('project.cfg')
        # no committed project config
        self.assertRaises(errors.BzrError,
            Project._check_control_branch, wt.branch)
        #
        wt.commit('commit and check')
        Project._check_control_branch(wt.branch)


class TestProjectLocal(TestCaseWithTransport):

    def setUp(self):
        TestCaseWithTransport.setUp(self)
        INSTALL_CMD()

    def test_initialize(self):
        t = get_transport('.')
        Project.initialize(t.base)
        self.assertIsDirectory('.scmproj', t)
        self.assertIsDirectory('.scmproj/.bzr', t)
        self.assertIsDirectory('.scmproj/alts', t)
        self.failUnlessExists(['.scmproj/project.cfg'])
        # check we have valid branch there
        branch.Branch.open('.scmproj')
        # you can't create twice
        self.assertRaises(errors.BzrError, Project.initialize, t.base)

    def test_error_wo_open(self):
        Project.initialize('.')
        p = Project('.')
        self.assertRaises(errors.BzrError, p.get_alt_name)
        self.assertRaises(errors.BzrError, p.get_subset_name)

    def test_open_w_subprojects(self):
        # we should not have warnings when there is no subprojects
        Project.initialize('.')

        p2 = Project('.')
        sio = StringIO()
        log_memento = trace.push_log_file(sio, log_format="%(message)s")
        try:
            p2.open()
        finally:
            trace.pop_log_file(log_memento)
        self.assertEqual('', sio.getvalue())

        # now with subproject section
        f = open(project.CONFIG_FN, "ab")
        try:
            f.write("\n[subproject kwa]\n")
        finally:
            f.close()

        p3 = Project('.')
        sio = StringIO()
        log_memento = trace.push_log_file(sio, log_format="%(message)s")
        try:
            p3.open()
        finally:
            trace.pop_log_file(log_memento)
        self.assertContainsRe(sio.getvalue(),
            'The scmproj plugin version .* \(from .*\) does not support subprojects yet')

    def test_info_smoke(self):
        Project.initialize('.')
        p2 = Project('.').open()
        sio = StringIO()
        p2.info(sio)
        self.assertNotEqual('', sio.getvalue())

    def test_clone(self):
        self.build_tree(['a/', 'b/'])
        Project.initialize('a')

        # existing project dir
        Project('a').clone('b')
        p2 = Project('b').open()

        # non-existing project dir
        Project('a').clone('c')
        p2 = Project('c').open()

    def test_clone_simple_branch(self):
        # we cannot clone if branch has no project.cfg
        self.make_branch('.scmproj')
        self.assertRaises(errors.BzrError, Project('.').clone, 'b')

    def test_run_actions(self):
        Project.initialize('.')

        f = open(project.CONFIG_FN, 'wb')
        try:
            f.write(reindentBlock("""\
                [PROJECT]
                VCS = bzr
                
                [COMPONENTS]
                BRANCH_URL = http://example.com/{BRANCH}
                
                [ALTS]
                dev = foo:bar
                test = foo:baz
                
                [component "foo"]
                RELPATH = .
                BRANCHES = trunk bar baz
                
                    [[branch "bar"]]
                    BRANCH_URL = http://another.net/branch
                    REVISION = xyz
                
                    [[branch "baz"]]
                    BRANCH_URL = https://host.org/project
                    REVISION = 1234
                
                [component "spam"]
                RELPATH = subdir
                BRANCHES = trunk
                    [[branch "trunk"]]
                    BRANCH_URL = bzr://server/branches/{BRANCH}
                """))
        finally:
            f.close()

        # actions for testing component iterator
        class LoggerAction(Action):

            def __call__(self, to_file, ko, kb, local_paths, local_urls):
                to_file.write(' ko.name = %s\n' % ko.name)
                to_file.write('  kb.name = %s' % kb.name)
                to_file.write('; kb.url = %s' % kb.url)
                to_file.write('; local_branch_name = %s' % os.path.basename(
                    local_paths['LOCAL_BRANCH']))
                to_file.write('\n')

        class DozingAction(Action):

            def __call__(self, to_file, ko, kb, local_paths, local_urls):
                to_file.write(' Z-z-z\n')

        action_list = [LoggerAction(), DozingAction()]

        # actually testing
        p2 = Project('.').open()
        p2.settings.workspace = 'boxed'

        # test simple
        sio = StringIO()
        p2.run_actions(action_list=action_list, to_file=sio)
        self.assertEquals(reindentBlock("""\
            # Component: foo
             ko.name = foo
              kb.name = trunk; kb.url = http://example.com/trunk; local_branch_name = trunk
             Z-z-z
            #------------------------------------------------------------
            # Component: spam
             ko.name = spam
              kb.name = trunk; kb.url = bzr://server/branches/trunk; local_branch_name = trunk
             Z-z-z
            #------------------------------------------------------------
            # Done.
            """), sio.getvalue())

        # test alt_name
        sio = StringIO()
        p2.run_actions(action_list=action_list, to_file=sio,
            alt_name='dev')
        self.assertEquals(reindentBlock("""\
            # Component: foo
             ko.name = foo
              kb.name = bar; kb.url = http://another.net/branch; local_branch_name = bar
             Z-z-z
            #------------------------------------------------------------
            # Component: spam
             ko.name = spam
              kb.name = trunk; kb.url = bzr://server/branches/trunk; local_branch_name = trunk
             Z-z-z
            #------------------------------------------------------------
            # Done.
            """), sio.getvalue())

        # test component list
        sio = StringIO()
        p2.run_actions(action_list=action_list, to_file=sio,
            alt_name='test', component_list=['foo'])
        self.assertEquals(reindentBlock("""\
            # Component: foo
             ko.name = foo
              kb.name = baz; kb.url = https://host.org/project; local_branch_name = baz
             Z-z-z
            #------------------------------------------------------------
            # Done.
            """), sio.getvalue())

        # test all_branches
        sio = StringIO()
        p2.run_actions(action_list=action_list, to_file=sio,
            alt_name='DEFAULT', all_branches=True)
        self.assertEquals(reindentBlock("""\
            # Component: foo
             ko.name = foo
              kb.name = trunk; kb.url = http://example.com/trunk; local_branch_name = trunk
             Z-z-z
             ko.name = foo
              kb.name = bar; kb.url = http://another.net/branch; local_branch_name = bar
             Z-z-z
             ko.name = foo
              kb.name = baz; kb.url = https://host.org/project; local_branch_name = baz
             Z-z-z
            #------------------------------------------------------------
            # Component: spam
             ko.name = spam
              kb.name = trunk; kb.url = bzr://server/branches/trunk; local_branch_name = trunk
             Z-z-z
            #------------------------------------------------------------
            # Done.
            """), sio.getvalue())

        # test cd_relpath
        sio = StringIO()
        self.assertRaises(OSError, p2.run_actions, action_list,
            to_file=sio, cd_relpath=True)

        os.mkdir('subdir')
        sio = StringIO()
        p2.run_actions(action_list=action_list, to_file=sio, cd_relpath=True)
        self.assertEquals(reindentBlock("""\
            # Component: foo
             ko.name = foo
              kb.name = trunk; kb.url = http://example.com/trunk; local_branch_name = trunk
             Z-z-z
            #------------------------------------------------------------
            # Component: spam
            #  Entering directory `subdir'
             ko.name = spam
              kb.name = trunk; kb.url = bzr://server/branches/trunk; local_branch_name = trunk
             Z-z-z
            #------------------------------------------------------------
            # Done.
            """), sio.getvalue())

    def test_fetch(self):
        wt = self.make_branch_and_tree('branch')
        self.build_tree(['branch/dir/', 'branch/file'])
        wt.add(['dir', 'file'])
        wt.commit('initial')

        Project.initialize('project')

        f = open(os.path.join('project', project.CONFIG_FN), 'wb')
        try:
            f.write(reindentBlock("""\
                [PROJECT]
                VCS = bzr

                [component "foo"]
                RELPATH = src
                URL =
                BRANCHES = trunk

                    [[branch "trunk"]]
                    BRANCH_URL = %s
                """) % urlutils.local_path_to_url(os.path.abspath('branch')))
        finally:
            f.close()

        p2 = Project('project').open()
        p2.settings.workspace = 'boxed'

        sio = StringIO()
        p2.fetch(alt_name='DEFAULT', to_file=sio)

        lines = sio.getvalue().splitlines()

        self.assertEquals(reindentBlock("""\
            # Component: foo
            #   Creating local shared repository foo...
            #   Creating local branch trunk...""").splitlines(),
            lines[:3])
        self.assertContainsRe(lines[3], r"  Pulling trunk from .*")
        self.assertEquals(reindentBlock("""\
            #------------------------------------------------------------
            # Done.""").splitlines(), lines[4:])

    def test_get_alt_name(self):
        Project.initialize('.')
        p = Project('.').open()
        self.assertEqual('DEFAULT', p.get_alt_name())
        p.settings.alt = 'foo'
        self.assertEqual('foo', p.get_alt_name())

    def test_get_subset_name(self):
        Project.initialize('.')
        p = Project('.').open()
        self.assertEqual('DEFAULT', p.get_subset_name())
        p.settings.subset = 'bar'
        self.assertEqual('bar', p.get_subset_name())

    def test_save_settings(self):
        Project.initialize('.')

        p = Project('.').open()
        self.assertEqual('DEFAULT', p.get_alt_name())
        self.assertEqual('DEFAULT', p.get_subset_name())

        p.save_settings(None, None)
        self.assertEqual('DEFAULT', p.get_alt_name())
        self.assertEqual('DEFAULT', p.get_subset_name())

        p.save_settings('foo', None)
        self.assertEqual('foo', p.get_alt_name())
        self.assertEqual('DEFAULT', p.get_subset_name())

        p.save_settings('boo', None)
        self.assertEqual('foo', p.get_alt_name())
        self.assertEqual('DEFAULT', p.get_subset_name())

        p.save_settings('boo', None, override=True)
        self.assertEqual('boo', p.get_alt_name())
        self.assertEqual('DEFAULT', p.get_subset_name())

        p.save_settings(None, 'spam')
        self.assertEqual('boo', p.get_alt_name())
        self.assertEqual('spam', p.get_subset_name())

        p.save_settings(None, 'eggs')
        self.assertEqual('boo', p.get_alt_name())
        self.assertEqual('spam', p.get_subset_name())

        p.save_settings(None, 'eggs', override=True)
        self.assertEqual('boo', p.get_alt_name())
        self.assertEqual('eggs', p.get_subset_name())


class TestProjectRemote(TestCaseWithSFTPServer):

    def setUp(self):
        TestCaseWithSFTPServer.setUp(self)
        INSTALL_CMD()

    def test_initialize(self):
        t_local = get_transport('server')
        sftp_url = self.get_url('server')
        Project.initialize(sftp_url)
        self.assertIsDirectory('.bzr', t_local)
        # check we have valid branch there as remote
        branch.Branch.open(sftp_url)
        # check we have valid branch there as local
        branch.Branch.open('server')
        # you can't create twice
        self.assertRaises(errors.BzrError, Project.initialize, sftp_url)

    def test_initialize_to_exisiting_dir(self):
        t_local = get_transport('.')
        Project.initialize(self.get_url())
        self.assertIsDirectory('.bzr', t_local)
        # check we have valid branch there as remote
        branch.Branch.open(self.get_url())
        # check we have valid branch there as local
        branch.Branch.open('.')
        # you can't create twice
        self.assertRaises(errors.BzrError, Project.initialize, self.get_url())

    def test_clone(self):
        self.build_tree(['a/', 'b/'])
        Project.initialize(self.get_url('a'))
        Project(self.get_url('a')).clone('b')
        p2 = Project('b').open()

    def test_clone_simple_branch(self):
        # we cannot clone if branch has no project.cfg
        self.make_branch('server')
        self.assertRaises(errors.BzrError, Project(self.get_url('server')).clone, 'b')

    def test_open(self):
        p = Project(self.get_url('.'))
        self.assertRaises(errors.BzrError, p.open)
