# Simple support for organizing vcs branches as complex project
#
# Copyright (C) 2008, 2009 Alexander Belchenko <bialix AT ukr net>
# License: GNU GPL v.2

"""VCS backends."""

from bzrlib.lazy_import import lazy_import
lazy_import(globals(), """
from bzrlib import (
    branch,
    errors,
    osutils,
    workingtree,
    )

from bzrlib.workingtree import WorkingTree
""")

from bzrlib.commands import run_bzr
from bzrlib.registry import Registry
from bzrlib.transport import get_transport


vcs_reg = Registry()


class Vcs(object):

    def is_shared_repo(self, url):
        raise NotImplementedError

    def create_repo(self, url, format=None):
        raise NotImplementedError

    def is_branch(self, url):
        raise NotImplementedError

    def create_branch(self, url, format=None, create_prefix=False):
        raise NotImplementedError

    def is_checkout(self, path):
        raise NotImplementedError

    def create_checkout(self, branch_location, to_location, light=None):
        raise NotImplementedError

    def pull(self, to_url, from_url, revision=None):
        raise NotImplementedError

    def push(self, to_url, from_url, revision=None):
        raise NotImplementedError

    def is_tree_changed(self, tree_path):
        """Quick check does working tree has changes."""
        raise NotImplementedError

    def commit_tree(self, tree_path, msgfile):
        """Commit working tree changes.
        @param  tree_path:  path to working tree.
        @param  msgfile:    path to file with commit message.
        """
        raise NotImplementedError

    def update_tree(self, tree_path):
        raise NotImplementedError

    def get_tip_revid(self, tree_url):
        raise NotImplementedError

    def add(self, tree_path, relpath):
        """Add file to the tree.
        @param  tree_path:  path to branch wth working tree.
        @param  relpath:    path of file to add, relative to tree_path.
        """
        raise NotImplementedError


class BzrVcs(Vcs):
    """Support for Bazaar branches and repositories."""

    def is_shared_repo(self, url):
        r_url = url + '/.bzr/repository'
        t = get_transport(r_url)
        return t.has('shared-storage')

    def create_repo(self, url, format=None):
        kmd = ['--no-aliases', '--builtin', 'init-repository', '-q', '--no-trees',
            url]
        if format:
            kmd.append('--format='+format)
        return run_bzr(kmd)

    def is_branch(self, url):
        b_url = url + '/.bzr/branch'
        t = get_transport(b_url)
        return t.has('format')

    def create_branch(self, url, format=None, create_prefix=False):
        kmd = ['--no-aliases', '--builtin', 'init', '-q', url]
        if format:
            kmd.append('--format='+format)
        if create_prefix:
            kmd.append('--create-prefix')
        return run_bzr(kmd)

    def is_checkout(self, path):
        fname = osutils.pathjoin(path, '.bzr', 'checkout', 'format')
        return osutils.isfile(fname)

    def _create_parent_transports(self, child_transport):
        """Create parent directories if needed."""
        parent_transport = child_transport.clone('..')
        pathlist = []
        for x in xrange(1, 50): #safetynet
            try:
                parent_transport.mkdir('.')
            except errors.NoSuchFile:
                # we need to go deeper
                pathlist.appen(child_transport)
                parent_transport = child_transport.clone('..')
            except errors.FileExists:
                break
            else:
                break
        for t in pathlist:
            t.mkdir('.')

    def create_checkout(self, branch_location, to_location, light=None):
        self._create_parent_transports(get_transport(to_location))
        kmd = ['--no-aliases', '--builtin', 'checkout', '-q',
            branch_location, to_location]
        if light:
            kmd.append('--lightweight')
        return run_bzr(kmd)

    def pull(self, to_url, from_url, revision=None):
        kmd = ['--no-aliases', '--builtin', 'pull', '-q',
            '-d', to_url, from_url]
        if revision:
            kmd.append('--revision='+revision)
        return run_bzr(kmd)

    def push(self, to_url, from_url, revision=None, use_existing_dir=False):
        kmd = ['--no-aliases', '--builtin', 'push', '-q',
            '-d', from_url, to_url]
        if revision:
            kmd.append('--revision='+revision)
        if use_existing_dir:
            kmd.append('--use-existing-dir')
        return run_bzr(kmd)

    def is_tree_changed(self, tree_path):
        tree = WorkingTree.open(tree_path)
        tree.lock_read()
        try:
            for changes in tree.iter_changes(tree.basis_tree()):
                # we need to filter out TREE_ROOT changes
                parent = changes[4]
                if parent == (None, None):  # filter out TREE_ROOT (?)
                    continue
                return True
            else:
                return False
        finally:
            tree.unlock()

    def commit_tree(self, tree_path, msgfile, quiet=False):
        kmd = ['--no-aliases', '--builtin', 'commit',
            tree_path, '--file', msgfile]
        if quiet:
            kmd.append('-q')
        return run_bzr(kmd)

    def update_tree(self, tree_path):
        kmd = ['--no-aliases', '--builtin', 'update', tree_path]
        return run_bzr(kmd)

    def get_tip_revid(self, tree_url):
        br = branch.Branch.open(tree_url)
        return br.last_revision()

    def add(self, tree_path, relpath):
        wt = workingtree.WorkingTree.open(tree_path)
        wt.add([relpath])


vcs_reg.register('bzr', BzrVcs)
vcs_reg.default_key = 'bzr'
